# -*- encoding: UTF-8 -*-
# Lightproof grammar checker for LibreOffice and OpenOffice.org
# http://launchpad.net/lightproof
# version 1.4.1 (2011-06-10)
#
# 2009-2011 (c) László Németh (nemeth at numbertext org), license: GNU LGPL

import uno
import unohelper
import re
from lightproof_impl_dicollecte_fr_FR import locales
from lightproof_impl_dicollecte_fr_FR import pkg
import lightproof_handler_dicollecte_fr_FR
from string import join

from com.sun.star.linguistic2 import XProofreader, XSupportedLocales
from com.sun.star.linguistic2 import ProofreadingResult, SingleProofreadingError
from com.sun.star.lang import XServiceInfo, XServiceName, XServiceDisplayName
from com.sun.star.lang import Locale
from com.sun.star.text.TextMarkupType import PROOFREADING

# loaded rules
langrule = {}
langrule2 = {}
# ignored rules
ignore = {}

# cache for morphogical analyses
analyses = {}
stems = {}
suggestions = {}

# assign Calc functions
calcfunc = None

# end of sentence regex ([.?!:;…]|.$)[ .?!…(nbsp)»”")]*
reEOS = re.compile(u"([.?!:;\u2026]|.$)[ .?!\u2026\u00a0\u00bb\u201d\u0022\u0029]*")
# begin of paragraph regex
reBOP = re.compile(u"^\W*")
# end of paragraph regex
reEOP = re.compile(u"\W*$")

# check settings
def option(lang, opt):
#    l = lang.Language + "_" + lang.Country
    l = "%s_FR" % lang.Language
    return lightproof_handler_dicollecte_fr_FR.get_option(l, opt)

# search and exclude morphologies
def morphex(rLoc, word, pattern, negpattern, noword=False):
    global analyses
    if not word:
        return noword
    if word not in analyses:
        x = spellchecker.spell(u"<?xml?><query type='analyze'><word>%s</word></query>"%word, rLoc, ())
        if not x:
            return None
        t = x.getAlternatives()
        if not t:
            t = ['']
        analyses[word] = t[0].split("</a>")[:-1]
    a = analyses[word]
    # check negative condition
    np = re.compile(negpattern)
    for s in a:
        res = np.search(s)
        if res:
            return False
    # search pattern
    p = re.compile(pattern)
    for s in a:
        res = p.search(s)
        if res:
            return True
    return False

# check pattern on stems
def stemchk(rLoc, word, pattern, strict=True, noword=False):
    global stems
    if not word:
        return noword
    if word not in stems:
        x = spellchecker.spell(u"<?xml?><query type='stem'><word>%s</word></query>"%word, rLoc, ())
        if not x:
            return False
        t = x.getAlternatives()
        if not t:
            t = ['']
        stems[word] = t
    a = stems[word]
    p = re.compile(pattern)
    if strict:
        for s in a:
            res = p.search(s)
            if not res:
                return False
        return True
    else:
        for s in a:
            res = p.search(s)
            if res:
                return True
        return False

# filtering affix fields (ds, is, ts etc.)
def onlymorph(st):
    if st != None:
        st = re.sub(r"^.*(st:|po:)", r"\1", st) # keep last word part
        st = re.sub(r"\b(?=[dit][sp]:)","@", st) # and its affixes
        st = re.sub(r"(?<!@)\b\w\w:\w+","", st).replace('@','').strip()
    return st

# if the pattern matches all analyses of the input word,
# return the last matched substring
def _morph(rLoc, word, pattern, onlyaffix, strict, noword):
    global analyses
    if not word:
        return noword
    if word not in analyses:
        x = spellchecker.spell(u"<?xml?><query type='analyze'><word>%s</word></query>"%word, rLoc, ())
        if not x:
            return None
        t = x.getAlternatives()
        if not t:
            t = [""]
        analyses[word] = t[0].split("</a>")[:-1]
    a = analyses[word]
    result = None
    p = re.compile(pattern)
    if strict :
        for i in a:
            if onlyaffix:
                i = onlymorph(i)
            result = p.search(i)
            if result:
                result = result.group(0)
            else:
                return None
        return result
    else:
        for i in a:
            if onlyaffix:
                i = onlymorph(i)
            result = p.search(i)
            if result:
                return result.group(0)
        return None

def morph(rLoc, word, pattern, strict=True, noword=False):
    return _morph(rLoc, word, pattern, False, strict, noword)

def affix(rLoc, word, pattern, strict=True, noword=False):
    return _morph(rLoc, word, pattern, True, strict, noword)

def spell(rLoc, word):
    if not word:
        return None
    return spellchecker.isValid(word, rLoc, ())

# get the tuple of the stem of the word or an empty array
def stem(rLoc, word):
    global stems
    if not word:
        return []
    if not word in stems:
        x = spellchecker.spell(u"<?xml?><query type='stem'><word>%s</word></query>"%word, rLoc, ())
        if not x:
            return []
        t = x.getAlternatives()
        if not t:
            t = []
        stems[word] = list(t)
    return stems[word]
    
# get the tuple of the morphological generation of a word or an empty array
def generate(rLoc, word, example):
    if not word:
        return []
    x = spellchecker.spell(u"<?xml?><query type='generate'><word>%s</word><word>%s</word></query>"%(word, example), rLoc, ())
    if not x:
        return []
    t = x.getAlternatives()
    if not t:
        t = []
    return list(t)
    
# get suggestions
def suggest(rLoc, word):
    global suggestions
    if not word:
        return None
    if word not in suggestions:
        x = spellchecker.spell("_" + word, rLoc, ())
        if not x:
            return word
        t = x.getAlternatives()
        suggestions[word] = join(t, "\n")
    return suggestions[word]

# get the nth word of the input string or None
def word(s, n):
    a = re.match("(?u)( [-.\w%]+){" + str(n-1) + "}( [-.\w%]+)", s)
    if not a:
        return ''
    return a.group(2)[1:]

# get the (-)nth word of the input string or None
def wordmin(s, n):
    a = re.search("(?u)([-.\w%]+ )([-.\w%]+ ){" + str(n-1) + "}$", s)
    if not a:
        return ''
    return a.group(1)[:-1]

def calc(funcname, par):
    global calcfunc
    global SMGR
    if calcfunc == None:
        calcfunc = SMGR.createInstance("com.sun.star.sheet.FunctionAccess")
        if calcfunc == None:
            return None
    return calcfunc.callFunction(funcname, par)

def proofread( nDocId, rText, rLoc, nStartOfSentencePos, nSuggestedSentenceEndPos, rProperties, bAnalyzeMode ):
    global ignore
    aErrs = []
    s = rText[nStartOfSentencePos:nSuggestedSentenceEndPos]
    for i in get_rule(rLoc, bAnalyzeMode):
        if i[0] and not str(i[0]) in ignore:
            for m in i[0].finditer(s):
                if not i[3] or eval(i[3]):
                    aErr = uno.createUnoStruct("com.sun.star.linguistic2.SingleProofreadingError" )
                    aErr.nErrorStart        = nStartOfSentencePos + m.start(0) # nStartOfSentencePos
                    aErr.nErrorLength       = m.end(0) - m.start(0)
                    aErr.nErrorType         = PROOFREADING
                    aErr.aRuleIdentifier    = str(i[0])
                    iscap = (i[4] and m.group(0)[0:1].isupper())
                    if i[1][0:1] == "=":
                        aErr.aSuggestions = tuple(cap(eval(i[1][1:]).split("\n"), iscap, rLoc))
                    elif i[1] == "_":
                        aErr.aSuggestions = ()
                    else:
                        aErr.aSuggestions = tuple(cap(m.expand(i[1]).split("\n"), iscap, rLoc))
                    aErr.aShortComment      = m.expand(i[2])
                    aErr.aFullComment       = m.expand(i[2])
                    aErr.aProperties        = ()
                    aErrs = aErrs + [aErr]
    return aErrs

def cap(a, iscap, rLoc):
    if iscap:
        for i in range(0, len(a)):
            if a[i][0:1] == "i":
                if rLoc.Language == "tr" or rLoc.Language == "az":
                    a[i] = u"\u0130" + a[i][1:]
                elif a[i][1:2] == "j" and rLoc.Language == "nl":
                    a[i] = "IJ" + a[i][2:]
                else:
                    a[i] = "I" + a[i][1:]
            else:
                a[i] = a[i].capitalize()
    return a

def get_module(module, bAnalyzeMode):
    try:
        if bAnalyzeMode:
            return langrule[module]
        return langrule2[module]
    except:
        if not module in locales:
            return None
        try:
            d = __import__("lightproof_" + pkg)
        except:
            print "Error: missing language data: " + module
            return None
        # compile regular expressions
        for i in d.dic:
            try:
                if re.compile("[(][?]iu[)]").match(i[0]):
                    i += [True]
                    i[0] = re.sub("[(][?]iu[)]", "(?u)", i[0])
                else:
                    i += [False]
                i[0] = re.compile(i[0])
            except:
                print "Lightproof: bad rule -- ", i[0]
                i[0] = None
        langrule[module] = d.dic
        for i in d.dic2:
            try:
                if re.compile("[(][?]iu[)]").match(i[0]):
                    i += [True]
                    i[0] = re.sub("[(][?]iu[)]", "(?u)", i[0])
                else:
                    i += [False]
                i[0] = re.compile(i[0])
            except:
                print "Lightproof: bad rule -- ", i[0]
                i[0] = None
        langrule2[module] = d.dic2
    if bAnalyzeMode:
        return langrule[module]
    return langrule2[module]

def get_rule(rLocale, bAnalyzeMode):
#    rules = get_module(rLocale.Language + "_" + rLocale.Country, bAnalyzeMode)
    rules = get_module("fr_FR", bAnalyzeMode)
    if not rules:
        rules = get_module(rLocale.Language, bAnalyzeMode)
    return rules


class Lightproof( unohelper.Base, XProofreader, XServiceInfo, XServiceName, XServiceDisplayName, XSupportedLocales):
    def __init__( self, ctx, *args ):
        global spellchecker
        global SMGR
        self.ctx = ctx
        self.ServiceName = "com.sun.star.linguistic2.Proofreader"
        self.ImplementationName = "org.openoffice.comp.pyuno.Lightproof." + pkg
        self.SupportedServiceNames = (self.ServiceName, )
        self.locales = []
        for i in locales:
            l = locales[i]
            self.locales += [Locale(l[0], l[1], l[2])]
        self.locales = tuple(self.locales)
        currentContext = uno.getComponentContext()
        SMGR = currentContext.ServiceManager
        spellchecker = SMGR.createInstanceWithContext("com.sun.star.linguistic2.SpellChecker", currentContext)
        lightproof_handler_dicollecte_fr_FR.load(currentContext)

    # XServiceName method implementations
    def getServiceName(self):
        return self.ImplementationName

    # XServiceInfo method implementations
    def getImplementationName (self):
        return self.ImplementationName

    def supportsService(self, ServiceName):
        return (ServiceName in self.SupportedServiceNames)

    def getSupportedServiceNames (self):
        return self.SupportedServiceNames

    # XSupportedLocales
    def hasLocale(self, aLocale):
        if aLocale in self.locales:
            return True
        for i in self.locales:
            if i.Country == "" and aLocale.Language == i.Language:
                return True
        return False

    def getLocales(self):
        return self.locales

    # XProofreader
    def isSpellChecker(self):
        return False

    def doProofreading(self, nDocId, rText, rLocale, nStartOfSentencePos, nSuggestedSentenceEndPos, rProperties):
        aRes = uno.createUnoStruct( "com.sun.star.linguistic2.ProofreadingResult" )
        aRes.aDocumentIdentifier = nDocId
        aRes.aText = rText
        aRes.aLocale = rLocale
        aRes.nStartOfSentencePosition = nStartOfSentencePos
        aRes.nStartOfNextSentencePosition = nSuggestedSentenceEndPos
        l = rText[nSuggestedSentenceEndPos:nSuggestedSentenceEndPos+1]
        while l == " ":
            aRes.nStartOfNextSentencePosition = aRes.nStartOfNextSentencePosition + 1
            l = rText[aRes.nStartOfNextSentencePosition:aRes.nStartOfNextSentencePosition+1]
        if aRes.nStartOfNextSentencePosition == nSuggestedSentenceEndPos and l!="":
            aRes.nStartOfNextSentencePosition = nSuggestedSentenceEndPos + 1
        aRes.nBehindEndOfSentencePosition = aRes.nStartOfNextSentencePosition
        if not option(rLocale, 'des'):
            aErrors = []
            # analyze by paragraph
            try:
                aErrors = proofread(nDocId, rText, rLocale, nStartOfSentencePos, aRes.nBehindEndOfSentencePosition, rProperties, True)
            except:
                pass
    	    
            # analyze by sentence
            lPos = []
            nBOS = reBOP.match(rText).end() # nStartOfSentencePos
            for m in reEOS.finditer(rText):
                lPos.append((nBOS, m.end()))
                nBOS = m.end()
            for pos in lPos:
                nl = pos[1] - pos[0]
                if nl > 4 and nl < 1500 :
                    try:
                        aErrors += proofread(nDocId, rText, rLocale, pos[0], pos[1], rProperties, False)
                    except:
                        pass
            
            aRes.aErrors = tuple(aErrors)
        else:
            aRes.aErrors = ()
        aRes.aProperties = ()
        aRes.xProofreader = self
        return aRes

    def ignoreRule(self, rid, aLocale):
        global ignore
        ignore[rid] = 1

    def resetIgnoreRules(self):
        global ignore
        ignore = {}

    # XServiceDisplayName
    def getServiceDisplayName(self, aLocale):
        return "Lightproof Grammar Checker (" + pkg + ")"

g_ImplementationHelper = unohelper.ImplementationHelper()
g_ImplementationHelper.addImplementation( Lightproof, "org.openoffice.comp.pyuno.Lightproof." + pkg, ("com.sun.star.linguistic2.Proofreader",),)

g_ImplementationHelper.addImplementation( lightproof_handler_dicollecte_fr_FR.LightproofOptionsEventHandler, "org.openoffice.comp.pyuno.LightproofOptionsEventHandler." + pkg, ("com.sun.star.awt.XContainerWindowEventHandler",),)

# in Compile.py, add the lines:
# s = re.sub(r"nword\\(\\s*(\\d)", r'nword(s[m.end():],\\1', s)	 # nword(n)
# s = re.sub(r"nword\\(\\s*-(\\d)", r'nwordmin(s[:m.start()],\\1', s) # nword(-n)

# get the nth word of the input string or None
def nword(s, n):
    a = re.match("(?u)( [ -.,'\u2019()\xab\xbb\u201c\u201d\\w%]+){" + str(n-1) + "}( [ -.,'\u2019()\xab\xbb\u201c\u201d\\w%]+)", s)
    if not a:
        return ''
    return a.group(2)[1:]

# get the (-)nth word of the input string or None
def nwordmin(s, n):
    a = re.search("(?u)([ -.,'\u2019()\xab\xbb\u201c\u201d\\w%]+ )([ -.,'\u2019()\xab\xbb\u201c\u201d\\w%]+ ){" + str(n-1) + "}$", s)
    if not a:
        return ''
    return a.group(1)[:-1]

